<div id="{{ figure_id }}">
  {{ background_svg }}
</div>
<div id="{{ figure_id }}_options">
  <button type="button" class="download btn btn-primary">Save JSON</button>
  <button type="button" class="reactionbutton btn btn-success ">Show Reaction Nodes</button>
  <button type="button" class="svgbutton btn btn-info ">Download SVG</button>
</div>
<script type="text/Javascript">
{{ javascript_source }}
</script>
